@@ -1,4 +1,5 @@
 import os
+import re
 from datetime import datetime
 from enum import Enum
 from typing import List
@@ -8,7 +9,7 @@ from fuzzywuzzy import fuzz
 from groclient import GroClient
 
 from grocropclient import crop_budgets
-from grocropclient.constants import FrequencyList, OtherItemsList, SourceList, MetricList, CropList, GRO_YIELD_MODELS, AreaWeightingSeriesList, ItemNameCategoryList, CROP_BUDGET_REGIONS, NATIONAL_AND_STATE_CROP_BUDGET_REGIONS, US_States, US_STATES_SET, US_COUNTIES_SET
+from grocropclient.constants import FrequencyList, OtherItemsList, SourceList, MetricList, CropList, GRO_YIELD_MODELS, AreaWeightingSeriesList, ItemNameCategoryList, CROP_BUDGET_REGIONS, NATIONAL_AND_STATE_CROP_BUDGET_REGIONS, US_States, US_STATES_SET, US_COUNTIES_SET, US_COUNTIES_FIPS_TO_REGION_ID, US_COUNTIES_REGION_ID_TO_FIPS
 
 
 API_HOST = "api.gro-intelligence.com"
@@ -598,7 +599,7 @@ class GroCropClient(object):
         """Returns all yield models as an array of {"id": …, "name": …}. """
         return [{"id": k, "name": "%s / %s" % (v["region_name"], v["item_name"])} for k,v, in GRO_YIELD_MODELS.items()]
 
-    def get_yield_model_data(self, yield_model_id, start_date=None):
+    def get_yield_model_data(self, yield_model_id: str, start_date: str = None, end_date: str = None):
         """Returns a given yield model, starting from a given data.
            The yield_model_id can be obtained from the result of calling get_all_yield_models().
         """
@@ -613,6 +614,9 @@ class GroCropClient(object):
                        "item_id":  ym["item_id"], "region_id": ym["region_id"], "metric_id": ym["metric_id"]}
         if start_date:
             params["start_date"] = start_date
+        if end_date:
+            params["end_date"] = end_date
+
         return self.get_data_points(**params)
 
     def get_yield_model_summary_last_N_days(self, yield_model_data, N:int = 5):
@@ -954,7 +958,7 @@ class GroCropClient(object):
     ## Climate ##
     #############
    
-    def get_Gro_Drought_Index(self, region_id, frequency = FrequencyList.DAILY.value, start_date=None):
+    def get_Gro_Drought_Index(self, region_id: int, frequency: int = FrequencyList.DAILY.value, start_date: str = None, end_date: str = None):
         """Returns Drought Index for a region and a frequency {daily, weekly, monthly} (use FrequencyList enum)."""
         if isinstance(frequency, Enum):
             frequency = frequency.value
@@ -964,63 +968,83 @@ class GroCropClient(object):
                    "region_id": region_id }
         if start_date:
             params["start_date"] = start_date
+        if end_date:
+            params["end_date"] = end_date
+
         return self.get_data_points(**params)
 
-    def get_Gro_Observed_Flood(self, region_id, start_date=None):
-        """Returns Gro Observed Flood for a given rgion."""
+    def get_Gro_Observed_Flood(self, region_id: int, start_date: str = None, end_date: str = None):
+        """Returns Gro Observed Flood for a given region."""
         params = { "metric_id": MetricList.OBSERVED_FLOOD, "item_id": OtherItemsList.WATER_AREAS,
-                   "frequency_id": FrequencyList.DAILY, "source_id": SourceList.GRO_DERIVED_GEOSPATIAL,
+                   "frequency_id": FrequencyList.DAILY.value, "source_id": SourceList.GRO_DERIVED_GEOSPATIAL,
                    "region_id": region_id }
         if start_date:
             params["start_date"] = start_date
+        if end_date:
+            params["end_date"] = end_date
+
         return self.get_data_points(**params)
 
-    def get_MODIS_NDVI(self, region_id, start_date=None):
+    def get_MODIS_NDVI(self, region_id: int, frequency: int = FrequencyList.EIGHT_DAY.value, start_date: str = None, end_date: str = None):
         params = { "metric_id": MetricList.VEGETATION, "item_id": OtherItemsList.VEGETATION_NDVI,
-                   "frequency_id": FrequencyList.EIGHT_DAY, "source_id": SourceList.GIMMS_MODIS,
+                   "frequency_id": frequency, "source_id": SourceList.GIMMS_MODIS,
                    "region_id": region_id }
         if start_date:
             params["start_date"] = start_date
+        if end_date:
+            params["end_date"] = end_date
+
         return self.get_data_points(**params)
 
-    def get_MODIS_LST(self, region_id, start_date=None):
+    def get_MODIS_LST(self, region_id: int, frequency: int = FrequencyList.DAILY.value, start_date: str = None, end_date: str = None):
         params = { "metric_id": MetricList.TEMPERATURE, "item_id": OtherItemsList.LAND_TEMPERATURE,
-                   "frequency_id": FrequencyList.DAILY, "source_id": SourceList.MODIS_TERRA,
+                   "frequency_id": frequency, "source_id": SourceList.MODIS_TERRA,
                    "region_id": region_id }
         if start_date:
             params["start_date"] = start_date
+        if end_date:
+            params["end_date"] = end_date
+
         return self.get_data_points(**params)
 
-    def get_ESA_SMOS(self, region_id, start_date=None):
+    def get_ESA_SMOS(self, region_id: int, frequency: int = FrequencyList.DAILY.value, start_date: str = None, end_date: str = None):
         params = { "metric_id": MetricList.AVAILABILITY_IN_SOIL, "item_id": OtherItemsList.SOIL_MOISTURE,
-                   "frequency_id": FrequencyList.DAILY, "source_id": SourceList.ESA_SMOS,
+                   "frequency_id": frequency, "source_id": SourceList.ESA_SMOS,
                    "region_id": region_id }
         if start_date:
             params["start_date"] = start_date
+        if end_date:
+            params["end_date"] = end_date
         return self.get_data_points(**params)
 
-    def get_GFS_Tmax(self, region_id, start_date=None):
+    def get_GFS_Tmax(self, region_id: int, start_date: str = None, end_date: str = None):
         params = { "metric_id": MetricList.TEMPERATURE, "item_id": OtherItemsList.TEMPERATURE_MAX,
-                   "frequency_id": FrequencyList.DAILY, "source_id": SourceList.GFS_00Z,
+                   "frequency_id": FrequencyList.DAILY.value, "source_id": SourceList.GFS_00Z,
                    "region_id": region_id, "show_revisions": True}
         if start_date:
             params["start_date"] = start_date
+        if end_date:
+            params["end_date"] = end_date
         return self.get_data_points(**params)
 
-    def get_GFS_Tmin(self, region_id, start_date=None):
+    def get_GFS_Tmin(self, region_id: int, start_date: str = None, end_date: str = None):
         params = { "metric_id": MetricList.TEMPERATURE, "item_id": OtherItemsList.TEMPERATURE_MIN,
-                   "frequency_id": FrequencyList.DAILY, "source_id": SourceList.GFS_00Z,
+                   "frequency_id": FrequencyList.DAILY.value, "source_id": SourceList.GFS_00Z,
                    "region_id": region_id, "show_revisions": True}
         if start_date:
             params["start_date"] = start_date
+        if end_date:
+            params["end_date"] = end_date
         return self.get_data_points(**params)
 
-    def get_GFS_Precip(self, region_id, start_date=None):
+    def get_GFS_Precip(self, region_id: int, start_date: str = None, end_date: str = None):
         params = { "metric_id": MetricList.PRECIPITATION_QUANTITY, "item_id": OtherItemsList.RAINFALL,
-                   "frequency_id": FrequencyList.DAILY, "source_id": SourceList.GFS_00Z,
+                   "frequency_id": FrequencyList.DAILY.value, "source_id": SourceList.GFS_00Z,
                    "region_id": region_id, "show_revisions": True}
         if start_date:
             params["start_date"] = start_date
+        if end_date:
+            params["end_date"] = end_date
         return self.get_data_points(**params)
 
     ####################
@@ -1090,3 +1114,94 @@ class GroCropClient(object):
                                                      region_id = region_id, 
                                                      method = method, 
                                                      latest_date_only = latest_date_only)
+
+    def reverse_geocode_points(self, points: list):
+        """
+        Takes a list of lat/long pairs and return a list of corresponding Gro regions
+
+        Parameters
+        ----------
+        points: list[list[float]]
+        The list of points we want to reverse geocode. It is a list of points, where each point is a list with the lat/long pair
+        Example: `[[33.4484, -112.0740], [42.3314, -83.0458], [-8.8742, 125.7275]]`
+
+        Returns 
+        -------
+        A list of Python dictionaries, where each dictionary contains details for one of the passed points. The order is the same as the order of the passed points:
+
+        ```
+        [{'latitude': 33.4484, 
+          'longitude': -112.074, 
+          'l5_id': 136859, 
+          'l5_name': 'Maricopa', 
+          'l4_id': 13053, 
+          'l4_name': 'Arizona', 
+          'l3_id': 1215, 
+          'l3_name': 'United States'}, 
+         ...,
+         {'latitude': -8.8742, 
+          'longitude': 125.7275,
+          'l5_id': 134452,
+          'l5_name': 'Turiscai',
+          'l4_id': 12774,
+          'l4_name': 'Manufahi', 
+          'l3_id': 1199, 
+          'l3_name': 'East Timor'}]
+
+        ```
+        """
+        return self._client.reverse_geocode_points(points)
+
+    def convert_fips_codes_to_region_ids(self, fips_codes: List[str]) -> List[int]:
+        """
+        Takes a list of FIPS codes and returns a list of corresponding Gro region IDs
+
+        Parameters
+        ----------
+        fips_codes: list[str]
+        The list of FIPS codes we want to convert to Gro region_ids. Each FIPS code should be passed as a string containing only the FIPS number
+        Example: `['19043', '19017']`
+
+        Returns 
+        -------
+        A list of integers, where each integer is a district-level Gro region ID for a provided FIPS code. The order is the same as the order of the passed FIPS codes. If there is no match for a provided FIPS code, we return `None`:
+
+        ```
+        [137566, 137553]
+        ```
+        """
+ 
+        output_region_ids: list = []
+
+        for f in fips_codes:
+            assert re.match("^[0-9]+$", f), f"FIPS codes must be passed as strings containing only the numeric FIPS code, but {f} was passed"
+            output_region_ids.append(US_COUNTIES_FIPS_TO_REGION_ID.get(f, None))
+
+        return output_region_ids
+
+    def convert_region_ids_to_fips_codes(self, region_ids: List[int]) -> List[str]:
+        """
+        Takes a list of Gro region IDs and returns a list of corresponding FIPS codes
+
+        Parameters
+        ----------
+        region_ids: list[int]
+        The list of Gro region_ids we want to convert to FIPS codes. Each region_id should be passed as an integer
+        Example: [137566, 137553]
+
+        Returns 
+        -------
+        A list of FIPS codes, where each FIPS is a string containing only the numeric code. The order is the same as the order of the passed region_ids. If there is no match for a provided region_id, we return `None`:
+
+        ```
+        ['19043', '19017']
+        ```
+        """
+ 
+        output_region_ids: list = []
+
+        for r in region_ids:
+            assert type(r)==int, f"region_ids must be passed as integers, but {r} was passed with type {type(r)}"
+            output_region_ids.append(US_COUNTIES_REGION_ID_TO_FIPS.get(r, None))
+
+        return output_region_ids
